/*
 * Copyright 2009 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Class KeyInfoArrayCache caches KeyInfo array objects in a singleton cache.
 */
public class KeyInfoArrayCache {
    private static Logger log = Logger.getLogger(KeyInfoArrayCache.class.getName());
	private static KeyInfoArrayCache instance = null;
	private Map<Class<?>, KeyInfo[]> cache;
	private int found;
	private int total;
	private KeyInfoArrayCache() {
		reset();
	}
	/**
	 * Get singleton instance.
	 * @return instance
	 */
	public synchronized static KeyInfoArrayCache getInstance() {
		if(instance == null) {
			instance = new KeyInfoArrayCache();
		}
		return instance;
	}
	/**
	 * Get KeyInfo array from cache.
	 * @param clz class we want KeyInfo array for
	 * @return KeyInfo array
	 */
	public synchronized KeyInfo[] get(Class<?> clz) {
		KeyInfo[] res = cache.get(clz);
		if(res != null) {
			found++;
            log.finer(clz.getName() + " found in KeyInfoArrayCache");
        } else {
            log.finer(clz.getName() + " not found in KeyInfoArrayCache");
        }
		total++;
		return res;
	}
	/**
	 * Put KeyInfo array into cache.
	 * @param clz class we have KeyInfo array for
	 * @param kia KeyInfo array
	 */
	public synchronized void put(Class<?> clz, KeyInfo[] kia) {
		cache.put(clz, kia);
        log.finer(clz.getName() + " put in KeyInfoArrayCache");
	}
	/**
	 * Get cache hit rate.
	 * @return hit rate
	 */
	public synchronized double getHitRate() {
		return (total > 0) ? (double)found/(double)total : -1;
	}
	/**
	 * Reset cache.
	 */
	public synchronized void reset() {
		cache = new HashMap<Class<?>, KeyInfo[]>();
		found = 0;
		total = 0;
        log.finer("KeyInfoArrayCache reset");
	}
	/**
	 * Convenience method to get KeyInfo array from cache and analyze class if not in cache.
	 * @param <T> type of record
	 * @param t class
	 * @return KeyInfo array
	 * @throws IsamException if problem with key
	 */
	public static <T> KeyInfo[] analyze(Class<T> t) throws IsamException {
		KeyInfoArrayCache cache = KeyInfoArrayCache.getInstance();
		KeyInfo[] keys = cache.get(t);
		if(keys == null) {
			keys = KeyInfo.analyze(t);
			cache.put(t, keys);
		}
		return keys;
	}
}
