/*
 * Copyright 2009 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.cache;

import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.isam.Key;
import dk.vajhoej.isam.KeyInfo;
import dk.vajhoej.isam.KeyInfoArrayCache;
import dk.vajhoej.record.RecordException;

/**
 * Class CacheIsamSource is a wrapper adding a cache to an ISAM source.
 */
public class CacheIsamSource implements IsamSource {
    private static Logger log = Logger.getLogger(CacheIsamSource.class.getName());
	private IsamSource real;
	private IsamCache cache;
	/**
	 * Create instance of CacheIsamSource with default cache implementation.
	 * @param real ISAM source
	 */
	public CacheIsamSource(IsamSource real) {
		this.real = real;
		this.cache = new StandardIsamCache();
	}
	/**
	 * Create instance of CacheIsamSource.
	 * @param real ISAM source
	 * @param cache cache implementation
	 */
	public CacheIsamSource(IsamSource real, IsamCache cache) {
		this.real = real;
		this.cache = cache;
	}
	public <T, TK extends Comparable<TK>> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException {
		T o = cache.get(t, k);
		if(o == null) {
            log.finer(t.getName() + " " + k + " not found in cache");
			o = real.read(t, k);
			if(o != null) {
			    multiPut(o);
                log.finer(t.getName() + " " + k + " put in cache");
			}
		} else {
            log.finer(t.getName() + " " + k + " found in cache");
		}
		return o;
	}
	public <T> void create(T o) throws IsamException, RecordException {
        real.create(o);
		multiPut(o);
	}
    public <T> void update(T o) throws IsamException, RecordException {
        real.update(o);
        multiPut(o);
    }
    public <T, TK extends Comparable<TK>> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        real.delete(t, k);
        T o = cache.get(t, k);
        if(o != null) {
            multiRemove(o);
        }
    }
	public void close() throws IsamException {
		real.close();
	}
	public <T,TK extends Comparable<TK>> IsamResult<T> readStart(Class<T> t) throws IsamException, RecordException {
		return real.readStart(t);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readGE(Class<T> t, Key<TK> k) throws IsamException, RecordException {
		return real.readGE(t, k);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readGT(Class<T> t, Key<TK> k) throws IsamException, RecordException {
		return real.readGT(t, k);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readLE(Class<T> t, Key<TK> k) throws IsamException, RecordException {
		return real.readLE(t, k);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readLT(Class<T> t, Key<TK> k) throws IsamException, RecordException {
		return real.readLT(t, k);
	}
	/**
	 * Get hit rate.
	 * @return cache hit rate
	 */
	public double getHitRate() {
		return cache.getHitRate();
	}
    private <T> void multiPut(T o) throws IsamException, RecordException {
        try {
            KeyInfo[] keys = KeyInfoArrayCache.analyze(o.getClass());
            for(int i = 0; i < keys.length; i++) {
                @SuppressWarnings({ "unchecked", "rawtypes" })
				Key<?> k = Key.create(i, (Comparable)keys[i].getField().get(o));
                cache.put(k, o);
                log.finer(o.getClass().getName() + " " + k + " put in cache");
            }
        } catch (IllegalAccessException e) {
            throw new RecordException("Cannot access field", e);
        }
    }
    private <T> void multiRemove(T o) throws IsamException, RecordException {
        try {
            KeyInfo[] keys = KeyInfoArrayCache.analyze(o.getClass());
            for(int i = 0; i < keys.length; i++) {
                @SuppressWarnings({ "unchecked", "rawtypes" })
				Key<?> k = Key.create(i, (Comparable)keys[i].getField().get(o));
                cache.remove(o.getClass(), k);
                log.finer(o.getClass().getName() + " " + k + " removed from cache");
            }
        } catch (IllegalAccessException e) {
            throw new RecordException("Cannot access field", e);
        }
    }
}
