/*
 * Copyright 2009 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.local;

import java.lang.reflect.Field;
import java.util.logging.Level;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.Key;
import dk.vajhoej.isam.KeyInfo;
import dk.vajhoej.isam.KeyInfoArrayCache;
import dk.vajhoej.record.FastStructReader;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.StructReader;

/**
 * Class LocalIsamResultReverse is a result from a find operation in a local ISAM source
 * for implementations that does not support reverse.
 */
public class LocalIsamResultReverse<T, TK extends Comparable<TK>> implements IsamResult<T> {
    private static Logger log = Logger.getLogger(LocalIsamResultReverse.class.getName());
    private Class<T> t;
    private Key<TK> k;
    private NativeIsam del;
    private byte[] buf;
    private boolean firstfnd;
    private boolean first;
    private Field f;
    /**
     * Create instance of LocalIsamResult.
     * @param t type to load
     * @param k key to find
     * @param del delegate to NativeIsam
     * @param buf buffer to receive data
     * @param firstfnd whether first record was found
     * @throws IsamException if problem
     */
    public LocalIsamResultReverse(Class<T> t, Key<TK> k, NativeIsam del, byte[] buf, boolean firstfnd) throws IsamException {
        this.t = t;
        this.k = k;
        this.del = del;
        this.buf = buf;
        this.firstfnd = firstfnd;
        first = true;
        KeyInfo[] keys = KeyInfoArrayCache.analyze(t);
        f = keys[k.getIx()].getField();
        f.setAccessible(true);
    }
    public boolean read() throws IsamException, RecordException {
        log.fine("Result read");
        if(first) {
            first = false;
            return firstfnd;
        } else {
            log.finer("Reverse lookup by key " + k);
            int stat = del.find(k.getIx(), k.getValue(), NativeIsam.COMP_LT, buf);
            return stat > 0;
        }
    }
    @SuppressWarnings("unchecked")
    public T current() throws IsamException, RecordException {
        log.fine("Result current");
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
        }
        T o;
 		if(LocalIsamSource.useFastStruct) {
			FastStructReader<T> fsr = FastSupport.getReader(t);
			fsr.init(buf);
	       o = fsr.read(t);
		} else {
	        StructReader sr = new StructReader(buf);
	        o = sr.read(t);
		}
        try {
            k = new Key<TK>(k.getIx(), (TK)f.get(o));
        } catch (IllegalArgumentException e) {
            throw new IsamException("Problem reversing");
        } catch (IllegalAccessException e) {
            throw new IsamException("Problem reversing");
        }
        return o;
    }
    public void close() throws IsamException {
        log.fine("Result close");
        if(!del.supportFindMixedOther()) {
        	del.close();
        }
    }
}
