/*
 * Copyright 2009 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.local;

import java.lang.reflect.Field;
import java.util.logging.Level;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamResult;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.isam.Key;
import dk.vajhoej.isam.Key0;
import dk.vajhoej.isam.KeyInfo;
import dk.vajhoej.isam.KeyInfoArrayCache;
import dk.vajhoej.record.FastStructReader;
import dk.vajhoej.record.FastStructWriter;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.StructInfo;
import dk.vajhoej.record.StructInfoCache;
import dk.vajhoej.record.StructReader;
import dk.vajhoej.record.StructWriter;

/**
 * Class LocalIsamSource is accessing a local ISAM file using native API.
 */
public class LocalIsamSource implements IsamSource {
	static boolean useFastStruct = false;
	static {
		String opt = System.getProperty("use.faststruct");
		if(opt != null && opt.equals("true")) useFastStruct = true;
	}
    private static Logger log = Logger.getLogger(LocalIsamSource.class.getName());
    private String fnm;
    private String implementation;
    private NativeIsam del;
    private <T, TK extends Comparable<TK>> byte[] fill(Class<T> t, Key<TK> k) throws IsamException, RecordException  {
    	try {
			T o = (T)t.newInstance();
			Field f = KeyInfoArrayCache.analyze(t)[k.getIx()].getField();
			f.setAccessible(true);
			f.set(o, k.getValue());
			if(useFastStruct) {
				FastStructWriter<T> fsw = FastSupport.getWriter(t);
				fsw.write(o);
				return fsw.getBytes();
			} else {
				StructWriter sw = new StructWriter();
				sw.write(o);
				return sw.getBytes();
			}
		} catch (InstantiationException e) {
			throw new IsamException(e);
		} catch (IllegalAccessException e) {
			throw new IsamException(e);
		}
    }
	private <T, TK extends Comparable<TK>> void fill(Class<T> t, Key<TK> k, byte[] buf) throws IsamException, RecordException  {
		System.arraycopy(fill(t, k), 0, buf, 0, buf.length);
	}
    /**
     * Create instance of LocaleIsamSource.
     * @param fnm filename
     * @param implementation ISAM implementation class
     * @param readonly true=&gt;only read, false=&gt;read+create+update+delete
     * @throws IsamException if error creating instance of ISAM implementation class
     */
	public LocalIsamSource(String fnm, String implementation, boolean readonly) throws IsamException {
	    this.fnm = fnm;
	    this.implementation = implementation;
	    try {
            del = (NativeIsam)Class.forName(implementation).newInstance();
        } catch (InstantiationException e) {
            throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
        } catch (IllegalAccessException e) {
            throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
        } catch (ClassNotFoundException e) {
            throw new IsamException("Error creating native ISAM implementation - class not found: " + implementation);
        }
		del.open(fnm, readonly);
        log.fine(fnm + " " + implementation + " initialized");
	}
	public <T, TK extends Comparable<TK>> T read(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Read class " + t.getName() + " with key " + k);
        StructInfo si = StructInfoCache.analyze(t);
        int len = si.getLength();
        if(!si.isFixedLength()) {
            len += 32768;
        }
        log.finer("Reading with buffer length " + len);
        byte[] buf = new byte[len];
        int stat;
        if(del.supportReadWithKey()) {
        	stat = del.read(k.getIx(), k.getValue(), buf);
        } else {
        	fill(t, k, buf);
			stat = del.read(buf);
        }
	    if(stat <= 0) {
            log.fine("Not found");
	        return null;
	    } else {
            log.fine("Found");
	    }
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf, 16));
        }
		if(useFastStruct) {
			FastStructReader<T> fsr = FastSupport.getReader(t);
			fsr.init(buf);
	        return fsr.read(t);
		} else {
	        StructReader sr = new StructReader(buf);
	        return sr.read(t);
		}
	}
	public <T> void create(T o) throws IsamException, RecordException {
        log.fine("Create class " + o.getClass().getName() + " with value " + o);
        byte[] buf;
		if(useFastStruct) {
			@SuppressWarnings("unchecked")
			FastStructWriter<T> fsw = FastSupport.getWriter((Class<T>)o.getClass());
			fsw.write(o);
			buf = fsw.getBytes();
		} else {
			StructWriter sw = new StructWriter();
			sw.write(o);
			buf = sw.getBytes();
		}
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        int stat;
        if(del.supportCreateWithKey()) {
            KeyInfo[] keys = KeyInfoArrayCache.analyze(o.getClass());
            Field f = keys[0].getField();
            f.setAccessible(true);
            try {
                Object k = f.get(o);
                log.fine("Key found " + k);
                stat = del.create(buf, 0, k);
            } catch (IllegalArgumentException e) {
                throw new IsamException("Problem creating");
            } catch (IllegalAccessException e) {
                throw new IsamException("Problem creating");
            }
        } else {
        	KeyInfoArrayCache.analyze(o.getClass()); // even though we are not using the key we want to check if it is valid
        	stat = del.create(buf);
        }
        if(stat != buf.length) {
            log.fine("Failure");
            throw new IsamException("Error creating record in file");
        } else {
            log.fine("Success");
        }
	}
    public <T> void update(T o) throws IsamException, RecordException {
        log.fine("Update class " + o.getClass().getName() + " with value " + o);
        byte[] buf;
		if(useFastStruct) {
			@SuppressWarnings("unchecked")
			FastStructWriter<T> fsw = FastSupport.getWriter((Class<T>)o.getClass());
			fsw.write(o);
			buf = fsw.getBytes();
		} else {
			StructWriter sw = new StructWriter();
			sw.write(o);
			buf = sw.getBytes();
		}
        log.finest("Length: " + buf.length);
        if(log.isLoggable(Level.FINEST)) {
            log.finest("Bytes:" + LogHelper.byteArrayToString(buf));
        }
        int stat;
        if(del.supportUpdateWithKey()) {
            KeyInfo[] keys = KeyInfoArrayCache.analyze(o.getClass());
            Field f = keys[0].getField();
            f.setAccessible(true);
            try {
                Object k = f.get(o);
                log.fine("Key found " + k);
                stat = del.update(buf, 0, k);
            } catch (IllegalArgumentException e) {
                throw new IsamException("Problem updating");
            } catch (IllegalAccessException e) {
                throw new IsamException("Problem updating");
            }
        } else {
        	KeyInfoArrayCache.analyze(o.getClass()); // even though we are not using the key we want to check if it is valid
            stat = del.update(buf);
        }
        if(stat != buf.length) {
            log.fine("Failure");
            throw new IsamException("Error updating record in file");
        } else {
            log.fine("Success");
        }
    }
    public <T, TK extends Comparable<TK>> void delete(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("Delete class " + t.getName() + " with key " + k);
    	KeyInfoArrayCache.analyze(t); // even though we are not using the key we want to check if it is valid
        int stat;
        if(del.supportDeleteWithKey()) {
        	stat = del.delete(k.getIx(), k.getValue());
        } else {
        	byte[] buf = fill(t, k);
        	stat = del.delete(buf);
        }
        if(stat <= 0) {
            log.fine("Failure");
            throw new IsamException("Error deleting record from file");
        } else {
            log.fine("Success");
        }
    }
	public void close() throws IsamException {
	    del.close();
        log.fine("Closed");
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readStart(Class<T> t) throws IsamException, RecordException {
        log.fine("ReadStart class " + t.getName() );
        Class<?> ktyp = KeyInfoArrayCache.analyze(t)[0].getField().getType();
        Key<TK> k;
        if(ktyp.equals(String.class)) {
        	@SuppressWarnings("unchecked")
			TK emptystring = (TK)" ";
			k = new Key0<TK>(emptystring);
        } else if(ktyp.equals(int.class) || ktyp.equals(short.class) || ktyp.equals(byte.class) ||
        		ktyp.equals(Integer.class) || ktyp.equals(Short.class) || ktyp.equals(Byte.class)) {
            @SuppressWarnings("unchecked")
			TK zero = (TK)(Integer)0;
			k = new Key0<TK>(zero);
        } else if(ktyp.equals(long.class)|| ktyp.equals(Long.class) ) {
			@SuppressWarnings("unchecked")
			TK zero = (TK)(Long)0L;
			k = new Key0<TK>(zero);
        } else {
        	throw new IsamException("Unsupported key type: " + ktyp.getName());
        }
        return readGE(t, k);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readGE(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadGE class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_GE);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readGT(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadGT class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_GT);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readLE(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadLE class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_LE);
	}
	public <T, TK extends Comparable<TK>> IsamResult<T> readLT(Class<T> t, Key<TK> k) throws IsamException, RecordException {
        log.fine("ReadLT class " + t.getName() + " with key " + k);
        return readXX(t, k, NativeIsam.COMP_LT);
	}
	private <T, TK extends Comparable<TK>> IsamResult<T> readXX(Class<T> t, Key<TK> k, int comp) throws IsamException, RecordException {
        NativeIsam del2;
        if(!del.supportFindMixedOther()) {
            try {
                del2 = (NativeIsam)Class.forName(implementation).newInstance();
            } catch (InstantiationException e) {
                throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
            } catch (IllegalAccessException e) {
                throw new IsamException("Error creating native ISAM implementation - problem calling constructor for: " + implementation);
            } catch (ClassNotFoundException e) {
                throw new IsamException("Error creating native ISAM implementation - class not found: " + implementation);
            }
            del2.open(fnm, true);
            log.fine(fnm + " " + implementation + " initialized");
        } else {
        	del2 = del;
        }
        StructInfo si = StructInfoCache.analyze(t);
        int len = si.getLength();
        if(!si.isFixedLength()) {
            len += 32768;
        }
        log.finer("Finding with buffer length " + len);
        byte[] buf = new byte[len];
        int stat;
        if(del2.supportFindWithKey()) {
        	stat = del2.find(k.getIx(), k.getValue(), comp, buf);
        } else {
        	fill(t, k, buf);
        	stat = del2.find(comp, buf);
        }
	    if(comp == NativeIsam.COMP_GE || comp == NativeIsam.COMP_GT || del.supportReverse()) {
            return new LocalIsamResult<T>(t, del2, buf, stat > 0);
    	} else {
            return new LocalIsamResultReverse<T, TK>(t, k, del2, buf, stat > 0);
    	}	    
	}
}
