/*
 * Copyright 2023 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.map;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.isam.KeyInfo;
import dk.vajhoej.isam.KeyInfoArrayCache;

/**
 * Class MapBase is a common base class for classes that provide a Map view of an isam file.
 *
 * @param <K> key type
 * @param <R> record type
 */
public abstract class MapBase<K extends Comparable<K>,R> implements Map<K,R> {
	private static Logger log = Logger.getLogger(MapBase.class.getName());
	/**
	 * Class IsamEntry is an entry in IsamMap.
	 * @param <EK> entry key type
	 * @param <EV> entry value type
	 */
	public static class IsamEntry<EK,EV> implements Entry<EK,EV> {
		private EK key;
		private EV value;
		/**
		 * Construct instance of IsamEntry.
		 * @param key key
		 * @param value value
		 */
		public IsamEntry(EK key, EV value) {
			super();
			this.key = key;
			this.value = value;
		}
		/**
		 * Get key.
		 * @return key
		 */
		public EK getKey() {
			return key;
		}
		/**
		 * Set key.
		 * @param key key
		 */
		public void setKey(EK key) {
			this.key = key;
		}
		/**
		 * Get value.
		 * @return value
		 */
		public EV getValue() {
			return value;
		}
		/**
		 * Set value.
		 * @param value value
		 */
		public EV setValue(EV value) {
			this.value = value;
			return value;
		}
	}
	protected static interface RecordHandler<K,R> {
		boolean process(R rec);
		Object result();
	}
	protected IsamSource src;
	protected Class<R> recordClass;
	protected int keyNameToIndex(String keynam) throws IsamException {
		KeyInfo[] ki = KeyInfoArrayCache.analyze(recordClass);
		for(int i = 0; i < ki.length; i++) {
			if(ki[i].getField().getName().equals(keynam)) {
				return i;
			}
		}
		throw new IsamException("Unknown key: " + keynam);
	}
	protected abstract void all(RecordHandler<K,R> h);
	private Field f = null;
	protected K getKey(R rec) throws IsamException {
		return getKey(rec, 0);
	}
	protected K getKey(R rec, int ix) throws IsamException {
		try {
			if(f == null) {
				f = KeyInfoArrayCache.analyze(recordClass)[ix].getField();
				f.setAccessible(true);
			}
			@SuppressWarnings("unchecked")
			K key = (K)(f.get(rec));
			return key;
		} catch (IllegalArgumentException e) {
			throw new IsamException("Error getting key", e);
		} catch (IllegalAccessException e) {
			throw new IsamException("Error getting key", e);
		}
	}
	// **** Map methods ****
	@Override
	public int size() {
		RecordHandler<K,R> h = new RecordHandler<K,R>() {
			int n = 0;
			@Override
			public boolean process(R rec) {
				n++;
				return true;
			}
			@Override
			public Object result() {
				return n;
			}
		};
		all(h);
		return (Integer)h.result();
	}
	@Override
	public boolean containsValue(final Object value) {
		RecordHandler<K,R> h = new RecordHandler<K,R>() {
			private boolean found = false;
			@Override
			public boolean process(R rec) {
				found = rec.equals(value);
				return !found;
			}
			@Override
			public Object result() {
				return found;
			}
		};
		all(h);
		return (Boolean)h.result();
	}
	@Override
	public Set<K> keySet() {
		RecordHandler<K,R> h = new RecordHandler<K,R>() {
			Set<K> keys = new LinkedHashSet<K>();
			@Override
			public boolean process(R rec) {
				K key;
				try {
					key = getKey(rec);
				} catch (IsamException e) {
					key = null;
				}
				keys.add(key);
				return true;
			}
			@Override
			public Object result() {
				return keys;
			}
		};
		all(h);
		@SuppressWarnings("unchecked")
		Set<K> res = (Set<K>)h.result();
		log.finest("keySet() returned " + res);
		return res;
	}
	@Override
	public Collection<R> values() {
		RecordHandler<K,R> h = new RecordHandler<K,R>() {
			List<R> values = new ArrayList<R>();
			@Override
			public boolean process(R rec) {
				values.add(rec);
				return true;
			}
			@Override
			public Object result() {
				return values;
			}
		};
		all(h);
		@SuppressWarnings("unchecked")
		List<R> res = (List<R>)h.result();
		log.finest("values() returned " + res);
		return res;
	}
	@Override
	public Set<Entry<K,R>> entrySet() {
		RecordHandler<K,R> h = new RecordHandler<K,R>() {
			Set<IsamEntry<K,R>> entries = new LinkedHashSet<IsamEntry<K,R>>();
			@Override
			public boolean process(R rec) {
				K key;
				try {
					key = getKey(rec);
				} catch (IsamException e) {
					key = null;
				}
				entries.add(new IsamEntry<K,R>(key, rec));
				return true;
			}
			@Override
			public Object result() {
				return entries;
			}
		};
		all(h);
		@SuppressWarnings("unchecked")
		Set<Entry<K,R>> res = (Set<Entry<K,R>>)h.result();
		log.finest("entrySet() returned " + res);
		return res;
	}
}
