/*
 * Copyright 2023 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.map;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.record.MapHelper;
import dk.vajhoej.record.RecordException;

/**
 * Class PyIsamMap is a Jython friendly wrapper around IsamMap.
 *
 * @param <K> key type
 * @param <R> record type
 */
public class PyIsamMap<K extends Comparable<K>,R> extends IsamMap<K,R> {
	private static Logger log = Logger.getLogger(PyIsamMap.class.getName());
	// **** Python friendly version ****
	/**
	 * Class PyKeyDispatcher operate on selected lookup key.
	 *
	 */
	public class PyKeyDispatcher extends KeyDispatcher {
		/**
		 * Create instance of PyKeyDispatcher.
		 * @param keyIndex key index
		 */
		public PyKeyDispatcher(int keyIndex) {
			super(keyIndex);
			log.fine("PyKeyDispatcher for " + keyIndex + " created");
		}
		/**
		 * Select records where lookup key is value.
		 * @param key lookup key
		 * @return PyIsamSubMap instance
		 */
		public <K2 extends Comparable<K2>> PyIsamSubMap<K2,R> is(K2 key) {
			log.fine("PyIsamSubMap for is " + key + " created");
			return new PyIsamSubMap<K2,R>(recordClass, src, keyIndex, key, key);
		}
		/**
		 * Select records where lookup key is value.
		 * @param key lookup key
		 * @return PyIsamSubMap instance
		 */
		public PyIsamSubMap<Integer,R> isInt(int key) {
			log.fine("PyIsamSubMap for is " + key + " created");
			return new PyIsamSubMap<Integer,R>(recordClass, src, keyIndex, key, key);
		}
		/**
		 * Select records where lookup key is value.
		 * @param key lookup key
		 * @return PyIsamSubMap instance
		 */
		public PyIsamSubMap<String,R> isString(String key) {
			log.fine("PyIsamSubMap for is " + key + " created");
			return new PyIsamSubMap<String,R>(recordClass, src, keyIndex, key, key);
		}
		/**
		 * Select records where lookup key is in interval.
		 * @param startKey start lookup key interval
		 * @param endKey end lookup key interval
		 * @return PyIsamSubMap instance
		 */
		public <K2 extends Comparable<K2>> PyIsamSubMap<K2,R> between(K2 startKey, K2 endKey) {
			log.fine("PyIsamSubMap for between " + startKey + " " + endKey + " created");
			return new PyIsamSubMap<K2,R>(recordClass, src, keyIndex, startKey, endKey);
		}
		/**
		 * Select records where lookup key is in interval.
		 * @param startKey start lookup key interval
		 * @param endKey end lookup key interval
		 * @return PyIsamSubMap instance
		 */
		public PyIsamSubMap<Integer,R> betweenInt(int startKey, int endKey) {
			log.fine("PyIsamSubMap for between " + startKey + " " + endKey + " created");
			return new PyIsamSubMap<Integer,R>(recordClass, src, keyIndex, startKey, endKey);
		}
		/**
		 * Select records where lookup key is in interval.
		 * @param startKey start lookup key interval
		 * @param endKey end lookup key interval
		 * @return PyIsamSubMap instance
		 */
		public PyIsamSubMap<String,R> betweenString(String startKey, String endKey) {
			log.fine("PyIsamSubMap for between " + startKey + " " + endKey + " created");
			return new PyIsamSubMap<String,R>(recordClass, src, keyIndex, startKey, endKey);
		}
	}
	// **** required by inheritage ****
	/**
	 * Create instance of PyIsampMap for RMS.
	 * @param <K> key type
	 * @param <R> record type
	 * @param fnm file name
	 * @param recordClass record class (template)
	 * @return PyIsamMap wrapper
	 * @throws IsamException if problems
	 */
	public static <K extends Comparable<K>,R> PyIsamMap<K,R> createIsamMapRMS(String fnm, Class<R> recordClass) throws IsamException {
		return new PyIsamMap<K,R>(fnm, "dk.vajhoej.vms.rms.IndexSequential", recordClass);
	}
	/**
	 * Create instance of PyIsampMap for BDB.
	 * @param <K> key type
	 * @param <R> record type
	 * @param fnm file name
	 * @param recordClass record class (template)
	 * @return PyIsamMap wrapper
	 * @throws IsamException if problems
	 */
	public static <K extends Comparable<K>,R> PyIsamMap<K,R> createIsamMapBDB(String fnm, Class<R> recordClass) throws IsamException {
		return new PyIsamMap<K,R>(fnm, "dk.vajhoej.bdb.BDB", recordClass);
	}
	/**
	 * Create instance of PyIsampMap for VBISAM.
	 * @param <K> key type
	 * @param <R> record type
	 * @param fnm file name
	 * @param recordClass record class (template)
	 * @return PyIsamMap wrapper
	 * @throws IsamException if problems
	 */
	public static <K extends Comparable<K>,R> IsamMap<K,R> createIsamMapVBISAM(String fnm, Class<R> recordClass) throws IsamException {
		return new PyIsamMap<K,R>(fnm, "dk.vajhoej.vbisam.VBISAM", recordClass);
	}
	/**
	 * Create instance of PyIsampMap.
	 * @param fnm file name
	 * @param implementation IsamSource implementation class name
	 * @param recordClass record class (template)
	 * @throws IsamException if problems
	 */
	public PyIsamMap(String fnm, String implementation, Class<R> recordClass) throws IsamException {
		super(fnm, implementation, recordClass);
		log.fine("Instance for " + fnm + " " + implementation + " " + recordClass + " instantiated");
	}
	/**
	 * Create instance of PyIsampMap.
	 * @param src IsamSource instance
	 * @param recordClass record class (template)
	 */
	public PyIsamMap(IsamSource src, Class<R> recordClass) {
		super(src, recordClass);
		log.fine("Instance for " + recordClass + " instantiated");
	}
	// **** provide compatibility with Python dictionary ****
	/**
	 * Jython friendly alias for keySet().
	 * @return key set
	 */
	public Set<K> keys() {
		Set<K> res = keySet();
		log.finest("keys() returning " + res);
		return res;
	}
	/**
	 * Jython friendly wrapper around entrySet().
	 * @return list of entries
	 */
	public List<List<Object>> items() {
		List<List<Object>> res = new ArrayList<List<Object>>();
		for(Entry<K,R> e : entrySet()) {
			List<Object> ee = new ArrayList<Object>();
			ee.add(e.getKey());
			ee.add(e.getValue());
			res.add(ee);
		}
		log.finest("items() returning " + res);
		return res;
	}
	// **** convenience ****
	/**
	 * Put record using key from record.
	 * @param value record
	 * @return if success value else null
	 */
	public R put(R value) {
		try {
			return put(getKey(value), value);
		} catch (IsamException e) {
			log.finer("put of " + value + " returned null");
			return null;
		}
	}
	/**
	 * Put dictionary. Dictionary get mapped to record. 
	 * @param dict dictionary
	 * @return if success value else null
	 */
	public R putDict(Map<String,Object> dict) {
		try {
			R o = MapHelper.fromMap(recordClass, dict);
			return put(o);
		} catch (RecordException e) {
			log.finer("put of " + dict + " returned null");
			return null;
		}
	}
	// **** override ****
	@Override
	public PyKeyDispatcher key() {
		return new PyKeyDispatcher(0);
	}
	@Override
	public PyKeyDispatcher key(int keyix) {
		return new PyKeyDispatcher(keyix);
	}
	/**
	 * Select lookup key.
	 * @param keyix key index
	 * @return key dispatcher
	 */
	public PyKeyDispatcher keyIndex(int keyix) {
		return key(keyix);
	}
	/**
	 * Select lookup key.
	 * @param keynam key name
	 * @return key dispatcher
	 */
	public PyKeyDispatcher keyName(String keynam) {
		try {
			return keyIndex(keyNameToIndex(keynam));
		} catch (IsamException e) {
			return null;
		}
	}
}
