/*
 * Copyright 2023 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.isam.test;

import static org.junit.Assert.*;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.Statement;
import java.util.List;
import java.util.Map;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import dk.vajhoej.isam.IsamException;
import dk.vajhoej.isam.IsamSource;
import dk.vajhoej.isam.emulated.EmulatedIsamSource;
import dk.vajhoej.isam.map.PyIsamMap;
import dk.vajhoej.isam.map.PyIsamSubMap;
import dk.vajhoej.record.RecordException;

public class TestPyIsamSubMap {
	private final static String DRIVER = "com.mysql.cj.jdbc.Driver";
	private final static String CONURL = "jdbc:mysql://localhost/Test";
	private final static String UN = "root";
	private final static String PW = "hemmeligt";
	@Before
	public void setup() {
		try {
			Class.forName(DRIVER);
			Connection con = DriverManager.getConnection(CONURL, UN, PW);
			Statement stmt = con.createStatement();
			stmt.executeUpdate("CREATE TABLE data (iv INTEGER NOT NULL, recblob BLOB, PRIMARY KEY(iv))");
			stmt.executeUpdate("CREATE TABLE multikeydata (k1 INTEGER NOT NULL, k2 INTEGER, k3 INTEGER, recblob BLOB, PRIMARY KEY(k1))");
			stmt.close();
			con.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	@After
	public void teardown() {
		try {
			Connection con = DriverManager.getConnection(CONURL, UN, PW);
			Statement stmt = con.createStatement();
			stmt.executeUpdate("DROP TABLE data");
			stmt.executeUpdate("DROP TABLE multikeydata");
			stmt.close();
			con.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	@Test
    public void testReadBetween() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            Map<Integer,Data> subm = m.key().between(258 + 2, 258 + 7);
            int i = 0;
            for(Data o2 : subm.values()) {
				assertEquals("iv", 258 + 2 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 6, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testPyReadBetween() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            PyIsamSubMap<Integer,Data> subm = (PyIsamSubMap<Integer,Data>)m.key().between(258 + 2, 258 + 7);
            int i = 0;
            for(List<Object> e : subm.items()) {
            	Data o2 = (Data)e.get(1);
				assertEquals("iv", 258 + 2 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 6, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadBetween0() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            Map<Integer,Data> subm = m.key(0).between(258 + 2, 258 + 7);
            int i = 0;
            for(Data o2 : subm.values()) {
				assertEquals("iv", 258 + 2 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 6, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testPyReadBetweenName() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            PyIsamSubMap<Integer,Data> subm = (PyIsamSubMap<Integer,Data>)m.keyName("iv").between(258 + 2, 258 + 7);
            int i = 0;
            for(List<Object> e : subm.items()) {
            	Data o2 = (Data)e.get(1);
				assertEquals("iv", 258 + 2 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 6, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testPyReadBetween0() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            PyIsamSubMap<Integer,Data> subm = (PyIsamSubMap<Integer,Data>)m.key(0).between(258 + 2, 258 + 7);
            int i = 0;
            for(List<Object> e : subm.items()) {
            	Data o2 = (Data)e.get(1);
				assertEquals("iv", 258 + 2 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 6, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadIs() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            Map<Integer,Data> subm = m.key().is(258 + 3);
            int i = 0;
            for(Data o2 : subm.values()) {
				assertEquals("iv", 258 + 3 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 1, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadIs0() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            PyIsamMap<Integer,Data> m = new PyIsamMap<Integer,Data>(isam, Data.class);
            for(int i = 0; i < 10; i++) {
				Data o = new Data();
				o.setIv(258 + i);
				o.setXv(123.456);
				o.setSv("ABC     ");
				m.put(o.getIv(), o);
            }
            Map<Integer,Data> subm = m.key(0).is(258 + 3);
            int i = 0;
            for(Data o2 : subm.values()) {
				assertEquals("iv", 258 + 3 + i, o2.getIv());
				assertEquals("xv", 123.456, o2.getXv(), 0.0005);
				assertEquals("sv", "ABC     ", o2.getSv());
				i++;
            }
            assertEquals("Number records", 1, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadBetweenMultiK1() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            Map<Integer,MultiKeyData> subm = m.key(0).between(5, 12);
            int i = 0;
            for(MultiKeyData o2 : subm.values() ) {
				assertEquals("k1", (4 + i) + 1, o2.getK1());
				assertEquals("k2", (4 + i) / 2 + 1, o2.getK2());
				assertEquals("k3", (4 + i) / 4 + 1, o2.getK3());
				i++;
            }
            assertEquals("Number records", 8, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testPyReadBetweenMultiK1() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            PyIsamSubMap<Integer,MultiKeyData> subm = m.key(0).between(5, 12);
            int i = 0;
            for(List<Object> e : subm.items()) {
            	MultiKeyData o2 = (MultiKeyData)e.get(1);
				assertEquals("k1", (4 + i) + 1, o2.getK1());
				assertEquals("k2", (4 + i) / 2 + 1, o2.getK2());
				assertEquals("k3", (4 + i) / 4 + 1, o2.getK3());
				i++;
            }
            assertEquals("Number records", 8, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadBetweenMultiK2() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            Map<Integer,MultiKeyData> subm = m.key(1).between(3, 6);
            int i = 0;
            for(MultiKeyData o2 : subm.values() ) {
				assertEquals("k1", (4 + i) + 1, o2.getK1());
				assertEquals("k2", (4 + i) / 2 + 1, o2.getK2());
				assertEquals("k3", (4 + i) / 4 + 1, o2.getK3());
				i++;
            }
            assertEquals("Number records", 8, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testPyReadBetweenMultiK2() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            PyIsamSubMap<Integer,MultiKeyData> subm = m.key(1).between(3, 6);
            int i = 0;
            for(List<Object> e : subm.items()) {
            	MultiKeyData o2 = (MultiKeyData)e.get(1);
				assertEquals("k1", (4 + i) + 1, o2.getK1());
				assertEquals("k2", (4 + i) / 2 + 1, o2.getK2());
				assertEquals("k3", (4 + i) / 4 + 1, o2.getK3());
				i++;
            }
            assertEquals("Number records", 8, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadBetweenMultiK3() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            Map<Integer,MultiKeyData> subm = m.key(2).between(2, 3);
            int i = 0;
            for(MultiKeyData o2 : subm.values() ) {
				assertEquals("k1", (4 + i) + 1, o2.getK1());
				assertEquals("k2", (4 + i) / 2 + 1, o2.getK2());
				assertEquals("k3", (4 + i) / 4 + 1, o2.getK3());
				i++;
            }
            assertEquals("Number records", 8, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testPyReadBetweenMultiK3() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            PyIsamSubMap<Integer,MultiKeyData> subm = m.key(2).between(2, 3);
            int i = 0;
            for(List<Object> e : subm.items()) {
            	MultiKeyData o2 = (MultiKeyData)e.get(1);
				assertEquals("k1", (4 + i) + 1, o2.getK1());
				assertEquals("k2", (4 + i) / 2 + 1, o2.getK2());
				assertEquals("k3", (4 + i) / 4 + 1, o2.getK3());
				i++;
            }
            assertEquals("Number records", 8, i);
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadIsMultiK1() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            Map<Integer,MultiKeyData> subm = m.key(0).is(2);
            assertEquals("Number records", 1, subm.size());
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadIsMultiK2() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            Map<Integer,MultiKeyData> subm = m.key(1).is(2);
            assertEquals("Number records", 2, subm.size());
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
	@Test
    public void testReadIsMultiK3() {
		try {
            IsamSource isam = new EmulatedIsamSource(DRIVER, CONURL, UN, PW);
            for(int i = 0; i < 16; i++) {
				MultiKeyData o = new MultiKeyData();
				o.setK1(i + 1);
				o.setK2(i / 2 + 1);
				o.setK3(i / 4 + 1);
				isam.create(o);
            }
            PyIsamMap<Integer,MultiKeyData> m = new PyIsamMap<Integer,MultiKeyData>(isam, MultiKeyData.class);
            Map<Integer,MultiKeyData> subm = m.key(2).is(2);
            assertEquals("Number records", 4, subm.size());
			isam.close();
		} catch (IsamException e) {
			fail("Unexpected exception: " + e);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
    }
}
